"""Fill in a module description here"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/utils.ipynb.

# %% auto 0
__all__ = ['get_default_data_root', 'hdf_files_from_path', 'write_dataset', 'write_array', 'iodata_to_hdf5', 'dataset_to_hdf5',
           'unzip_download', 'unrar_download', 'download']

# %% ../nbs/utils.ipynb 3
from nonlinear_benchmarks import *
from collections.abc import Iterator
from pathlib import Path
import numpy as np
import h5py
import os

# %% ../nbs/utils.ipynb 5
def get_default_data_root() -> Path:
    """
    Returns the default root directory for datasets.

    Checks the 'IDENTIBENCH_DATA_ROOT' environment variable first,
    otherwise defaults to '~/.identibench_data'.
    """
    return Path(os.environ.get('IDENTIBENCH_DATA_ROOT', Path.home() / '.identibench_data'))


# %% ../nbs/utils.ipynb 7
def _dummy_dataset_loader(
    save_path: Path, # Directory where the dummy dataset files will be written
    force_download: bool = False, # Argument for interface compatibility
    create_train_valid_dir: bool = False # If True, create a 'train_valid' subdir as well
    ) -> None:
    """Creates a dummy dataset structure with minimal HDF5 files for testing."""
    save_path = Path(save_path)
    if save_path.is_dir() and not force_download: return

    save_path.mkdir(parents=True, exist_ok=True)
    seq_len = 50
    subdirs = ['train', 'valid', 'test']
    if create_train_valid_dir:
        subdirs.append('train_valid')

    for subdir in subdirs:
        subdir_path = save_path / subdir
        subdir_path.mkdir(exist_ok=True)
        n_files = 1 if subdir == 'train_valid' else 2 # Create fewer files in train_valid for testing differentiation
        for i in range(n_files):
            dummy_file_path = subdir_path / f'{subdir}_{i}.hdf5'
            try:
                with h5py.File(dummy_file_path, 'w') as f:
                    f.create_dataset('u0', data=np.random.rand(seq_len).astype(np.float32))
                    f.create_dataset('u1', data=np.random.rand(seq_len).astype(np.float32))
                    f.create_dataset('y0', data=np.random.rand(seq_len).astype(np.float32))
                    f.attrs['fs'] = 10.0
            except Exception as e: print(f"Failed to create dummy file {dummy_file_path}: {e}")

# %% ../nbs/utils.ipynb 8
def hdf_files_from_path(fpath: Path) -> list[Path]:
        return sorted(list(fpath.glob('*.hdf5'))) 

# %% ../nbs/utils.ipynb 10
def _load_sequences_from_files(
        file_paths: list[Path],
        u_cols: list[str],
        y_cols: list[str],
        x_cols: list[str]|None = None,
        win_sz: int|None = None,
        stp_sz: int|None = None
    ) -> Iterator[tuple[np.ndarray, np.ndarray, np.ndarray|None]]:
    if not file_paths: return iter([]) 
    if win_sz is None and stp_sz is not None:
        raise ValueError("win_sz must be provided if stp_sz is provided")
    if stp_sz is None and win_sz is not None:
        raise ValueError("stp_sz must be provided if win_sz is provided")

    for file_path in file_paths:
        try:
            with h5py.File(file_path, 'r') as f:
                u_data = np.stack([f[col][()] for col in u_cols], axis=-1).astype(np.float32)
                y_data = np.stack([f[col][()] for col in y_cols], axis=-1).astype(np.float32)
                x_data = np.stack([f[col][()] for col in x_cols], axis=-1).astype(np.float32) if x_cols else None
        except Exception as e:
            print(f"Warning: Error reading {file_path}: {e}")
            continue

        if win_sz is None:
            yield u_data, y_data, x_data
            continue

        seq_len = u_data.shape[0]
        if seq_len < win_sz: continue

        for start in range(0, seq_len - win_sz + 1, stp_sz):
            end = start + win_sz
            x_win = x_data[start:end] if x_data is not None else None
            yield u_data[start:end], y_data[start:end], x_win

# %% ../nbs/utils.ipynb 14
def write_dataset(
     group: h5py.File | h5py.Group,
     ds_name: str,
     data: np.ndarray,
     dtype: str = 'f4',
     chunks: tuple[int, ...]|None = None
     ) -> None:
     group.create_dataset(ds_name, data=data, dtype=dtype, chunks=chunks)

# %% ../nbs/utils.ipynb 17
def write_array(
        group: h5py.File | h5py.Group,
        ds_name: str,
        data: np.ndarray,
        dtype: str = 'f4',
        chunks: tuple[int, ...]|None = None
    ) -> None:
    'Writes a 2d numpy array rowwise to a hdf5 file.'
    for i in range(data.shape[1]):
        write_dataset(group, f'{ds_name}{i}', data[:,i], dtype, chunks)

# %% ../nbs/utils.ipynb 19
def iodata_to_hdf5(iodata:Input_output_data, # data to save to file
            hdf_dir:Path, # Export directory for hdf5 files
            f_name:str = None # name of hdf5 file without '.hdf5' ending
            ) -> Path:
    data_2d = iodata.atleast_2d()
    u,y = data_2d.u, data_2d.y
    
    os.makedirs(hdf_dir,exist_ok=True)
    if f_name is None: f_name = iodata.name
    
    hdf_path = Path(hdf_dir) / f'{f_name}.hdf5'.replace(" ", "_")
    with h5py.File(hdf_path,'w') as f:
        write_array(f,'u',u)
        write_array(f,'y',y)
        
        # Save sampling_rate and init_window_size as attributes
        if iodata.sampling_time is not None:
            f.attrs['fs'] = 1/iodata.sampling_time
        if iodata.state_initialization_window_length is not None:
            f.attrs['init_sz'] = iodata.state_initialization_window_length

    return hdf_path 

# %% ../nbs/utils.ipynb 26
def dataset_to_hdf5(train:tuple, #tuple of Input_output_data for training
                    valid:tuple,#tuple of Input_output_data for validation
                    test:tuple,#tuple of Input_output_data for test
                    save_path: Path, #directory the files are written to, created if it does not exist
                    train_valid: tuple = None # optional tuple of unsplit Input_output_data for training and validation
                    ) -> None:
    'Save a dataset consisting of training, validation, and test set in hdf5 format in seperate subdirectories'
    save_path = Path(save_path)
    
    dict_data = {'train':train,
                 'valid':valid,
                 'test':test,
                 'train_valid':train_valid}
    for subset,ds_entries in dict_data.items():
        if ds_entries is None: continue
        if isinstance(ds_entries,tuple):
            if not isinstance(ds_entries[0],Input_output_data): raise ValueError(f'Data has to be stored in tuples of Input_output_data. Got {type(ds_entries[0])}')
        else:
            if not isinstance(ds_entries,Input_output_data): raise ValueError(f'Data has to be stored in Input_output_data. Got {type(ds_entries)}')
            dict_data[subset] = (ds_entries,)

    os.makedirs(save_path,exist_ok=True)

    for subset,ds_entries in dict_data.items():
        if ds_entries is None: continue
        for idx,iodata in enumerate(ds_entries):
            iodata_to_hdf5(iodata,save_path / subset,f'{subset}_{idx}')

# %% ../nbs/utils.ipynb 30
import requests
import io
import os
import zipfile
import rarfile

# %% ../nbs/utils.ipynb 31
def unzip_download(url:str, #url to file to download
                   extract_dir: Path = Path('.') #directory the archive is extracted to
                   ) -> None:
    'downloads a zip archive to ram and extracts it'
    response = requests.get(url)
    with zipfile.ZipFile(io.BytesIO(response.content)) as zip_ref:
        zip_ref.extractall(extract_dir)

# %% ../nbs/utils.ipynb 32
def unrar_download(url:str, #url to file to download
                   extract_dir: Path = Path('.') #directory the archive is extracted to
                   ) -> None:
    'downloads a rar archive to ram and extracts it'
    response = requests.get(url)
    with rarfile.RarFile(io.BytesIO(response.content)) as zip_ref:
        zip_ref.extractall(extract_dir)

# %% ../nbs/utils.ipynb 33
def download(url:str, #url to file to download
             target_dir: Path = Path('.')
             ) -> Path:
    fname = Path(url).name
    if os.path.isfile(fname): return
    response = requests.get(url)
    p_name = Path(target_dir).joinpath(fname)
    with open(p_name, "wb") as file:
        file.write(response.content)
    return p_name



"""Fill in a module description here"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/benchmark.ipynb.

# %% auto 0
__all__ = ['BenchmarkSpecSimulation', 'BenchmarkSpecPrediction', 'TrainingContext', 'run_benchmark', 'run_multiple_benchmarks']

# %% ../nbs/benchmark.ipynb 2
from pathlib import Path
from collections.abc import Iterator, Callable
from typing import Any
import random
import numpy as np
import time
from .utils import get_default_data_root,_load_sequences_from_files, hdf_files_from_path

# %% ../nbs/benchmark.ipynb 5
def aggregate_metric_score(test_results,metric_func, score_name=None ,sequence_aggregation_func=np.mean,window_aggregation_func=np.mean):
    # iterate over test_results and calculate metric score for each (y_pred,y_test) tuple, if prediction, iterate over nested tuples with nested loop
    if score_name is None:
        score_name = metric_func.__name__
    if isinstance(test_results[0], list):
        scores = []
        for windowed_sequence in test_results:
            scores.append(window_aggregation_func([metric_func(y_pred,y_test) for y_pred,y_test in windowed_sequence]))
    else:
        scores = [[metric_func(y_pred,y_test) for y_pred,y_test in test_results]]
    return {score_name: sequence_aggregation_func(scores)}

# %% ../nbs/benchmark.ipynb 6
class BenchmarkSpecBase: pass 
class BenchmarkSpecBase:
    """
    Base class for benchmark specifications, holding common attributes.
    """
    def __init__(self,
                 name: str, # Unique name identifying this benchmark task.
                 dataset_id: str, # Identifier for the raw dataset source.
                 u_cols: list[str], # list of column names for input signals (u).
                 y_cols: list[str], # list of column names for output signals (y).
                 metric_func: Callable[[np.ndarray, np.ndarray], float], # Primary metric: `func(y_true, y_pred)`.
                 x_cols: list[str]|None = None, # Optional state inputs (x).
                 sampling_time: float|None = None, # Optional sampling time (seconds).
                 download_func: Callable[[Path, bool], None]|None = None, # Dataset preparation func.
                 test_model_func: Callable[[BenchmarkSpecBase, Callable], dict[str, Any]] = None,
                 custom_test_evaluation = None, 
                 init_window: int|None = None, # Steps for warm-up, potentially ignored in evaluation.
                 data_root: [Path, Callable[[], Path]] = get_default_data_root # root dir for dataset, may be a callable or path
                ):
        self.name = name
        self.dataset_id = dataset_id
        self.u_cols = u_cols
        self.y_cols = y_cols
        self.metric_func = metric_func
        self.x_cols = x_cols
        self.sampling_time = sampling_time
        self.download_func = download_func
        self.test_model_func = test_model_func
        self.custom_test_evaluation = custom_test_evaluation
        self.init_window = init_window
        self._data_root = data_root

    @property
    def data_root(self) -> Path:
        """Returns the evaluated data root path."""
        if isinstance(self._data_root, Callable):
            return self._data_root()
        return self._data_root

    @property
    def dataset_path(self) -> Path:
        """Returns the full path to the dataset directory."""
        return self.data_root / self.dataset_id
    
    @property
    def test_path(self) -> Path:
        """Returns the full path to the directory containing the test files ."""
        return self.dataset_path / 'test'
    
    @property
    def train_path(self) -> Path:
        """Returns the full path to the directory containing the train files."""
        return self.dataset_path / 'train'
    
    @property
    def valid_path(self) -> Path:
        """Returns the full path to the directory containing the valid files."""
        return self.dataset_path / 'valid'
    
    @property
    def train_valid_path(self) -> Path:
        """Returns the full path to the directory containing the train_valid files."""
        return self.dataset_path / 'train_valid'
    
    @property
    def train_files(self) -> list[Path]:
        """Returns the list of hdf5 files in the training directory."""
        return hdf_files_from_path(self.train_path)
    
    @property
    def valid_files(self) -> list[Path]:
        """Returns the list of hdf5 files in the validation directory."""
        return hdf_files_from_path(self.valid_path)
    
    @property
    def train_valid_files(self) -> list[Path]:
        """Returns the list of hdf5 files in the train_valid directory if it exists, otherwise returns the union of the train and valid directories."""
        train_valid_files = hdf_files_from_path(self.train_valid_path)
        if train_valid_files: 
            return train_valid_files
        else:
            train_files = hdf_files_from_path(self.train_path)
            valid_files = hdf_files_from_path(self.valid_path)
            return sorted(train_files+valid_files)

    @property
    def test_files(self) -> list[Path]:
        """Returns the list of hdf5 files in the test directory."""
        return hdf_files_from_path(self.test_path)

    def ensure_dataset_exists(self, force_download: bool = False) -> None:
        """Checks if the dataset exists, downloads/prepares it if needed."""
        # (Implementation remains the same as before)
        dataset_path = self.dataset_path
        if self.download_func is None:
            print(f"Warning: No download function for '{self.name}'. Assuming data exists at {dataset_path}")
            if not dataset_path.is_dir():
                 print(f"Warning: Dataset directory {dataset_path} not found.")
            return

        if not dataset_path.is_dir() or force_download:
            print(f"Preparing dataset for '{self.name}' at {dataset_path}...")
            self.data_root.mkdir(parents=True, exist_ok=True)
            try:
                self.download_func(dataset_path, force_download)
                print(f"Dataset '{self.name}' prepared successfully.")
            except Exception as e:
                print(f"Error preparing dataset '{self.name}': {e}")
                raise

# %% ../nbs/benchmark.ipynb 7
def _test_simulation(specs, model):
    results = []
    for u_test, y_test, _ in _load_sequences_from_files(specs.test_files, specs.u_cols, specs.y_cols, specs.x_cols):
        y_pred = model(u_test,y_test[:specs.init_window])
        y_test_win = y_test[specs.init_window:]
        y_pred = y_pred[-y_test_win.shape[0]:]
        results.append((y_pred,y_test_win))
    return results

# %% ../nbs/benchmark.ipynb 8
class BenchmarkSpecSimulation(BenchmarkSpecBase):
    """
    Specification for a simulation benchmark task.

    Inherits common parameters from BaseBenchmarkSpec.
    Use this when the goal is to simulate the system's output given the input `u`.
    """
    def __init__(self,
                name: str, # Unique name identifying this benchmark task.
                dataset_id: str, # Identifier for the raw dataset source.
                u_cols: list[str], # list of column names for input signals (u).
                y_cols: list[str], # list of column names for output signals (y).
                metric_func: Callable[[np.ndarray, np.ndarray], float], # Primary metric: `func(y_true, y_pred)`.
                x_cols: list[str]|None = None, # Optional state inputs (x).
                sampling_time: float|None = None, # Optional sampling time (seconds).
                download_func: Callable[[Path, bool], None]|None = None, # Dataset preparation func.
                test_model_func: Callable[[BenchmarkSpecBase, Callable], dict[str, Any]] = _test_simulation,
                custom_test_evaluation = None, 
                init_window: int|None = None, # Steps for warm-up, potentially ignored in evaluation.
                data_root: [Path, Callable[[], Path]] = get_default_data_root # root dir for dataset, may be a callable or path
            ):
        self.name = name
        self.dataset_id = dataset_id
        self.u_cols = u_cols
        self.y_cols = y_cols
        self.metric_func = metric_func
        self.x_cols = x_cols
        self.sampling_time = sampling_time
        self.download_func = download_func
        self.test_model_func = test_model_func
        self.custom_test_evaluation = custom_test_evaluation
        self.init_window = init_window
        self._data_root = data_root

# %% ../nbs/benchmark.ipynb 9
def _test_prediction(specs, model):
    results = []
    for u_test, y_test, _ in _load_sequences_from_files(specs.test_files, specs.u_cols, specs.y_cols, specs.x_cols):
        #iterate through windows of u_test and y_test
        window_results = []
        for i in range(0, u_test.shape[0] - specs.init_window - specs.pred_horizon, specs.pred_step):
            u_test_win = u_test[i:i+specs.init_window+specs.pred_horizon]
            y_test_win = y_test[i:i+specs.init_window+specs.pred_horizon]
            y_pred = model(u_test_win,y_test_win[:specs.init_window])
            window_results.append((y_pred,y_test_win))
        results.append(window_results)
    return results

# %% ../nbs/benchmark.ipynb 10
class BenchmarkSpecPrediction(BenchmarkSpecBase):
     """
     Specification for a k-step ahead prediction benchmark task.

     Inherits common parameters from BaseBenchmarkSpec and adds prediction-specific ones.
     Use this when the goal is to predict `y` some steps ahead based on past `u` and `y`.
     """
     def __init__(self,
               name: str, # Unique name identifying this benchmark task.
               dataset_id: str, # Identifier for the raw dataset source.
               u_cols: list[str], # list of column names for input signals (u).
               y_cols: list[str], # list of column names for output signals (y).
               metric_func: Callable[[np.ndarray, np.ndarray], float], # Primary metric: `func(y_true, y_pred)`.
               pred_horizon: int, # The 'k' in k-step ahead prediction (mandatory for this type).
               pred_step: int, # Step size for k-step ahead prediction (e.g., predict y[t+k] using data up to t).
               x_cols: list[str]|None = None, # Optional state inputs (x).
               sampling_time: float|None = None, # Optional sampling time (seconds).
               download_func: Callable[[Path, bool], None]|None = None, # Dataset preparation func.
               test_model_func: Callable[[BenchmarkSpecBase, Callable], dict[str, Any]] = _test_prediction,
               custom_test_evaluation = None, 
               init_window: int|None = None, # Steps for warm-up, potentially ignored in evaluation.
               data_root: [Path, Callable[[], Path]] = get_default_data_root # root dir for dataset, may be a callable or path
          ):
          if pred_horizon <= 0:
               raise ValueError("pred_horizon must be a positive integer for PredictionBenchmarkSpec.")
          self.pred_horizon = pred_horizon
          self.pred_step = pred_step
          
          self.name = name
          self.dataset_id = dataset_id
          self.u_cols = u_cols
          self.y_cols = y_cols
          self.metric_func = metric_func
          self.x_cols = x_cols
          self.sampling_time = sampling_time
          self.download_func = download_func
          self.test_model_func = test_model_func
          self.custom_test_evaluation = custom_test_evaluation
          self.init_window = init_window
          self._data_root = data_root

# %% ../nbs/benchmark.ipynb 17
class TrainingContext:
    """
    Context object passed to the user's training function (`build_predictor`).

    Holds the benchmark specification, hyperparameters, and seed.
    Provides methods to access the raw, full-length training and validation data sequences.
    Windowing/batching for training must be handled within the user's `build_predictor` function.
    """
    # Explicit __init__ for nbdev documentation compatibility
    def __init__(self, 
                 spec: BenchmarkSpecBase, # The benchmark specification.
                 hyperparameters: dict[str, Any], # User-provided dictionary containing model and training hyperparameters.
                 seed: int|None = None # Optional random seed for reproducibility.
                ):
        # Standard attribute assignment
        self.spec = spec
        self.hyperparameters = hyperparameters
        self.seed = seed

    # --- Data Access Methods ---

    def get_train_sequences(self) -> Iterator[tuple[np.ndarray, np.ndarray, np.ndarray|None]]:
        """Returns a lazy iterator yielding raw (u, y, x) tuples for the 'train' subset."""
        return _load_sequences_from_files(
            file_paths=self.spec.train_files,
            u_cols=self.spec.u_cols,
            y_cols=self.spec.y_cols,
            x_cols=self.spec.x_cols,
        )

    def get_valid_sequences(self) -> Iterator[tuple[np.ndarray, np.ndarray, np.ndarray|None]]:
        """Returns a lazy iterator yielding raw (u, y, x) tuples for the 'valid' subset."""
        return _load_sequences_from_files(
            file_paths=self.spec.valid_files,
            u_cols=self.spec.u_cols,
            y_cols=self.spec.y_cols,
            x_cols=self.spec.x_cols,
        )

    def get_train_valid_sequences(self) -> Iterator[tuple[np.ndarray, np.ndarray, np.ndarray|None]]:
        """
        Returns a lazy iterator yielding raw (u, y, x) tuples for combined training and validation.

        Checks for a 'train_valid' subset directory first. If it exists, loads data from there.
        If not, it loads data from 'train' and 'valid' subsets sequentially.
        """
        return _load_sequences_from_files(
            file_paths=self.spec.train_valid_files,
            u_cols=self.spec.u_cols,
            y_cols=self.spec.y_cols,
            x_cols=self.spec.x_cols,
        )

# %% ../nbs/benchmark.ipynb 20
def run_benchmark(spec, build_model, hyperparameters={}, seed=None):

    if seed is None:
        seed = random.randint(0, 2**32 - 1)
    
    results = {
        'benchmark_name': spec.name,
        'dataset_id': spec.dataset_id,
        'hyperparameters': hyperparameters,
        'seed': seed,
        'training_time_seconds': np.nan,
        'test_time_seconds': np.nan,
        'benchmark_type' : type(spec).__name__,
        'metric_name': spec.metric_func.__name__,
        'metric_score': np.nan,
        'custom_scores': {},
        'model_predictions': []
    }

    spec.ensure_dataset_exists() 

    context = TrainingContext(spec=spec, hyperparameters=hyperparameters, seed=seed) 

    train_start_time = time.monotonic()
    model = build_model(context) 
    train_end_time = time.monotonic()
    results['training_time_seconds'] = train_end_time - train_start_time

    if model is None:
        raise RuntimeError(f"build_model for {spec.name} did not return a model.") 
        
    test_start_time = time.monotonic()
    test_results = spec.test_model_func(spec, model) # get list of (y_pred,y_test) tuples
    test_end_time = time.monotonic()
    results['test_time_seconds'] = test_end_time - test_start_time

    results['model_predictions'] = test_results

    results.update(aggregate_metric_score(test_results, spec.metric_func, score_name='metric_score'))
    if spec.custom_test_evaluation is not None:
        results['custom_scores'] = spec.custom_test_evaluation(test_results, spec)
        
    return results

# %% ../nbs/benchmark.ipynb 21
# Define a very simple build_model function for the example
def _dummy_build_model(context):
    print(f"Building model with spec: {context.spec.name}, seed: {context.seed}")

    def dummy_model(u_test,y_test):
        output_dim = len(context.spec.y_cols) 
        return np.zeros((u_test.shape[0], output_dim))
        
    return dummy_model # Return the callable model

# %% ../nbs/benchmark.ipynb 30
def run_multiple_benchmarks(
    specs: list[BenchmarkSpecBase] | dict[str, BenchmarkSpecBase], # Collection of specs to run
    build_model: Callable[[TrainingContext], Callable], # User function to build the model/predictor
    hyperparameters: dict[str, Any]|None = None, # Hyperparameters passed to build_model
    seed: int|None = None, # Base random seed
    continue_on_error: bool = True, # If True, continue running benchmarks even if one fails
) -> list[dict[str, Any]]:
    """
    Runs multiple benchmarks sequentially using the same build_model function.

    Args:
        specs: A list or dictionary containing the BenchmarkSpec objects to run.
        build_model: A callable that accepts a TrainingContext and returns a trained model/predictor function.
        hyperparameters: A dictionary of hyperparameters passed to the build_model function.
        seed: An optional integer seed passed to each run_benchmark call for reproducibility.
        continue_on_error: If True, catches exceptions during individual benchmark runs, prints a warning,
                           and continues. If False, stops on the first error.

    Returns:
        A list of result dictionaries containing the results from successful benchmark runs.
    """
    results_list = []
    hyperparameters = hyperparameters or {} # Ensure it's a dict

    # Determine the list of specification objects to iterate over
    spec_objects = list(specs.values()) if isinstance(specs, dict) else list(specs)

    print(f"--- Starting benchmark run for {len(spec_objects)} specifications ---")

    for i, spec in enumerate(spec_objects):
        spec_name = getattr(spec, 'name', f'Unnamed Spec {i+1}') # Get name for logging
        print(f"\n[{i+1}/{len(spec_objects)}] Running benchmark: {spec_name}")

        try:
            # Run the individual benchmark
            result = run_benchmark(
                spec=spec,
                build_model=build_model,
                hyperparameters=hyperparameters,
                seed=seed # Pass the same base seed to each run
            )
            results_list.append(result)
            print(f"  -> Success: {spec_name} completed.")

        except Exception as e:
            print(f"  -> ERROR running benchmark '{spec_name}': {e}")
            if not continue_on_error:
                print("Stopping due to error (continue_on_error=False).")
                raise 

    print(f"\n--- Benchmark run finished. {len(results_list)}/{len(spec_objects)} completed successfully. ---")
    return results_list



"""Fill in a module description here"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/datasets/quadrotor_pi.ipynb.

# %% auto 0
__all__ = ['quad_pi_u', 'quad_pi_x_v', 'quad_pi_x_q', 'quad_pi_x_w', 'quad_pi_x', 'quad_pi_y_vdot', 'quad_pi_y_wdot', 'quad_pi_y',
           'BenchmarkQuadPi_Simulation', 'BenchmarkQuadPi_Prediction', 'dl_quad_pi']

# %% ../../nbs/datasets/quadrotor_pi.ipynb 2
from nonlinear_benchmarks.utilities import get_tmp_benchmark_directory
import identibench.benchmark as idb
import identibench.metrics
from pathlib import Path
import os
import h5py
import numpy as np
import pandas as pd

import gdown
import bagpy
import glob
import scipy
import shutil
from scipy.signal import butter, lfilter, lfilter_zi

# %% ../../nbs/datasets/quadrotor_pi.ipynb 3
fnames_test = ['ovalz_10',
                'ovalz_4',
                '8z_5',
                '8z_6',
                'line8z_4',
                'wz_12',
                'v_8',
                'vT_5']

fnames_valid = ['oval_5',
                'linez_4',
                '8_5',
                '8_7',
                'w_12'
                'vz_7',
                'v_8']

def get_parent_dir(f_name: str # name of the flight
                  ):
    if f_name in fnames_valid:
        return 'valid'
    elif f_name in fnames_test:
        return 'test'
    else:
        return 'train'

# %% ../../nbs/datasets/quadrotor_pi.ipynb 4
def parseBag(topic, path):
  bag = bagpy.bagreader(path, verbose=False)
  return pd.read_csv(bag.message_by_topic(topic))

def shift(arr):
  if len(arr.shape) == 1:
    arr = arr.reshape((len(arr), 1))
  return np.vstack((np.nan * np.ones(shape=(1, arr.shape[1])), arr[:-1]))

def shiftFilteredSpline(sampled_data, step):
  for k in sampled_data.keys():
    if "filt" in k:
      sampled_data[k] = sampled_data[k][step:]
    else:
      sampled_data[k] = sampled_data[k][:-step]

def differentiate(arr, dt):
  a_dot = (arr - shift(arr)) / dt
  a_dot[0, :] = np.zeros((a_dot.shape[1]))
  return a_dot

def differentiateFivePointStencil(arr, dt):
  a_dot = [
    [np.zeros((arr.shape[1]))],
    [(arr[2] - arr[1]) / dt[1]]
  ]
  for i in range(2, len(arr) - 2):
    a_dot.append([(- arr[i + 2] + 8 * arr[i + 1] - 8 * arr[i - 1] + arr[i - 2]) * (1. / (12 * dt[i]))])
  a_dot.append([(arr[-2] - arr[-3]) / dt[-2]])
  a_dot.append([(arr[-1] - arr[-2]) / dt[-1]])
  return np.concatenate(a_dot, axis=0)

def dropNoise(arr, t, dt):
  drop_indeces = []
  for i in range(1, len(arr)):
    if np.any(abs(arr[i]) - abs(sum(arr[i-4:i])) > 0):
      drop_indeces.append(i)
  return np.delete(arr, drop_indeces, axis=0), np.delete(t, drop_indeces, axis=0), np.delete(dt, drop_indeces, axis=0)

def applySavitzkyGolayFilter(arr, window_length, poly_order):
  return np.array([savitzkyGolayFilter(arr[:, i], window_length, poly_order) for i in range(arr.shape[1])]).T

def savitzkyGolayFilter(data, window_length, poly_order):
  return scipy.signal.savgol_filter(data, window_length, poly_order)

def applyButterLowpassFilter(arr):
  return np.array([butterLowpassFilter(arr[:, i]) for i in range(arr.shape[1])]).T

def butterLowpassFilter(data):
  nyq = 0.5 * frequency
  normal_cutoff = cutoff / nyq
  b, a = butter(order, normal_cutoff, btype='lowpass', analog=False)
  zi = lfilter_zi(b, a)
  y = lfilter(b, a, data, zi=data[0] * zi)[0]
  return y

def appendHistory(df, data_columns, label_columns, history_length):
  state_columns = [col for col in data_columns if "f" not in col]
  df_state = df[state_columns]
  df_state_history = df_state.rename(columns={col: col + "_t" for col in state_columns})
  for j in range(1, 1 + history_length):
    shifted_df = df_state.shift(j)
    for k in range(j):
      shifted_df.iloc[k] = shifted_df.iloc[j]  # repeat initial elements where shift has left NaNs
    col_names = {col: col + "_t-" + str(j) for col in list(df.columns)}
    shifted_df.rename(columns=col_names, inplace=True)
    df_state_history = pd.concat([shifted_df, df_state_history], axis=1)

  input_columns = [col for col in data_columns if "f" in col]
  df_input = df[input_columns]
  df_input_history = df_input.rename(columns={col: col + "_t" for col in input_columns})
  for j in range(1, 1 + history_length):
    shifted_df = df_input.shift(j)
    for k in range(j):
      shifted_df.iloc[k] = shifted_df.iloc[j]  # repeat initial elements where shift has left NaNs
    col_names = {col: col + "_t-" + str(j) for col in list(df.columns)}
    shifted_df.rename(columns=col_names, inplace=True)
    df_input_history = pd.concat([shifted_df, df_input_history], axis=1)

  df_history = pd.concat([df_state_history, df_input_history, df[label_columns]], axis=1)
  return df_history

def computeSpline(str, arr, t, steps, cols):
  if len(t.shape) == 1:
    t = t.reshape((t.shape[0], 1))
  splines = {}
  for i in range(len(cols)):
    splines[str + "_" + cols[i]] = scipy.interpolate.CubicSpline(t[:, 0], arr[:, i])(steps, 0)
  return splines

def nominalModel(data, thrust_coeff, torque_coeff, inertia, mass, arm_length):
  q = np.vstack((np.vstack((data["q_w"], data["q_x"])), np.vstack((data["q_y"], data["q_z"])))).T
  f = np.vstack((np.vstack((data["f_0"], data["f_1"])), np.vstack((data["f_2"], data["f_3"])))).T
  w = np.vstack((np.vstack((data["w_x"], data["w_y"])), data["w_z"])).T

  vdot = []
  for i in range(q.shape[0]):
    thrust = f[i, 0] + f[i, 1] + f[i, 2] + f[i, 3]
    quat_norm = q[i, 0] ** 2 + q[i, 1] ** 2 + q[i, 2] ** 2 + q[i, 3] ** 2
    vdot.append([
      (1. / mass) * thrust * 2. * (q[i, 0] * q[i, 2] + q[i, 1] * q[i, 3]) / quat_norm,
      (1. / mass) * thrust * 2. * (q[i, 2] * q[i, 3] - q[i, 0] * q[i, 1]) / quat_norm,
      (1. / mass) * thrust * (1. - 2. * q[i, 1] * q[i, 1] - 2. * q[i, 2] * q[i, 2]) / quat_norm - 9.8066
    ])

  wdot = []
  km_kf = torque_coeff / thrust_coeff
  for i in range(w.shape[0]):
    wdot.append([
      (arm_length * (f[i, 0] + f[i, 1] - f[i, 2] - f[i, 3])  + inertia[1] * w[i, 1] * w[i, 2] - inertia[2] * w[i, 1] * w[i, 2]) / inertia[0],
      (arm_length * (-f[i, 0] + f[i, 1] + f[i, 2] - f[i, 3]) - inertia[0] * w[i, 0] * w[i, 2] + inertia[2] * w[i, 0] * w[i, 2]) / inertia[1],
      (km_kf      * (f[i, 0] - f[i, 1] + f[i, 2] - f[i, 3])  + inertia[0] * w[i, 0] * w[i, 1] - inertia[1] * w[i, 0] * w[i, 1]) / inertia[2]
    ])

  return np.array(vdot), np.array(wdot)

# low pass filter
frequency = 100.0
cutoff = 5
order = 4
shift_step = 9

def processBag(path):
  # quadrotor physics
  mass = 0.25
  thrust_coeff = 4.37900e-09
  torque_coeff = 3.97005e-11
  arm_length = 0.076
  inertia = np.array([0.000601, 0.000589, 0.001076])

  # load dataframes
  df_odom = parseBag('/dragonfly17/odom', str(path))
  df_imu = parseBag('/dragonfly17/imu', str(path))
  df_motor = parseBag('/dragonfly17/motor_rpm', str(path))

  # compute time
#   t_odom = df_odom.apply(lambda r: rospy.Time(r["header.stamp.secs"], r["header.stamp.nsecs"]).to_sec(), axis=1)
#   t_imu = df_imu.apply(lambda r: (r["header.stamp.secs"] + (r["header.stamp.nsecs"] / 1e9)), axis=1)
#   t_motor = df_motor.apply(lambda r: rospy.Time(r["header.stamp.secs"], r["header.stamp.nsecs"]).to_sec(), axis=1)

  t_odom = df_odom.apply(lambda r: r["header.stamp.secs"] + r["header.stamp.nsecs"] / 1e9, axis=1)
  t_imu = df_imu.apply(lambda r: r["header.stamp.secs"] + r["header.stamp.nsecs"] / 1e9, axis=1)
  t_motor = df_motor.apply(lambda r: r["header.stamp.secs"] + r["header.stamp.nsecs"] / 1e9, axis=1)

  t_odom = t_odom.to_numpy().reshape((len(t_odom), 1))
  t_imu = t_imu.to_numpy().reshape((len(t_imu), 1))
  t_motor = t_motor.to_numpy().reshape((len(t_motor), 1))
  dt_odom = t_odom - shift(t_odom)
  dt_imu = t_imu - shift(t_imu)
  dt_motor = t_motor - shift(t_motor)
  dt_odom[np.isnan(dt_odom)] = 0.
  dt_imu[np.isnan(dt_imu)] = 0.
  dt_motor[np.isnan(dt_motor)] = 0.

  # sampling steps
  sampling_bounds = [max(np.min(t_odom), np.min(t_imu), np.min(t_motor)),
                     min(np.max(t_odom), np.max(t_imu), np.max(t_motor))]
  sampling_bounds[0] = round(sampling_bounds[0] - sampling_bounds[0] % 1. / frequency, 4)
  sampling_bounds[1] = round(sampling_bounds[1] - sampling_bounds[1] % 1. / frequency, 4)
  sampling_steps = np.arange(sampling_bounds[0], sampling_bounds[1], 1. / frequency)[:-1]

  # store all processed data in a dictionary
  sampled_data = {"t": sampling_steps}

  ## position
  p = df_odom[["pose.pose.position.x", "pose.pose.position.y", "pose.pose.position.z"]].to_numpy()
  sampled_data.update(computeSpline("p", p, t_odom, sampling_steps, "xyz"))

  ## orientation
  q = df_odom[["pose.pose.orientation.w", "pose.pose.orientation.x",
               "pose.pose.orientation.y", "pose.pose.orientation.z"]].to_numpy()
  # r = [scipy_rotation.from_quat(q[i, :]).as_matrix() for i in range(q.shape[0])]
  q = applyButterLowpassFilter(q)
  sampled_data.update(computeSpline("q", q, t_odom, sampling_steps, "wxyz"))

  ## motor speeds
  u = df_motor[["rpm_0", "rpm_1", "rpm_2", "rpm_3"]].to_numpy()
  u, t_filt, _ = dropNoise(u, t_motor, dt_motor)
  u = applyButterLowpassFilter(u)
  sampled_data.update(computeSpline("u", u, t_filt, sampling_steps, "0123"))

  ## motor thrusts
  f = (u ** 2) * thrust_coeff
  sampled_data.update(computeSpline("f", f, t_filt, sampling_steps, "0123"))

  ## linear velocity
  v = df_odom[["twist.twist.linear.x", "twist.twist.linear.y", "twist.twist.linear.z"]].to_numpy()
  v, t_filt, dt_filt = dropNoise(v, t_odom, dt_odom)
  v = applyButterLowpassFilter(v)
  # v = applySavitzkyGolayFilter(v, window_length=101, poly_order=4)
  sampled_data.update(computeSpline("v", v, t_filt, sampling_steps, "xyz"))

  ## angular velocity
  # w = df_odom[["twist.twist.angular.x", "twist.twist.angular.y", "twist.twist.angular.z"]].to_numpy()
  w = df_imu[["angular_velocity.x", "angular_velocity.y", "angular_velocity.z"]].to_numpy()
  w = w * np.array([1, -1, -1])
  w = applyButterLowpassFilter(w)
  # w = applySavitzkyGolayFilter(w, window_length=101, poly_order=4)
  sampled_data.update(computeSpline("w", w, t_imu, sampling_steps, "xyz"))

  ## linear acceleration
  vdot = differentiate(v, dt_filt)
  # vdot = differentiateFivePointStencil(v, dt_filt)
  vdot = applyButterLowpassFilter(vdot)
  # vdot = applySavitzkyGolayFilter(vdot, window_length=101, poly_order=4)
  sampled_data.update(computeSpline("vdot", vdot, t_filt, sampling_steps, "xyz"))

  ## angular acceleration
  wdot = differentiate(w, dt_imu)
  # wdot = differentiateFivePointStencil(w, dt_imu)
  wdot = applyButterLowpassFilter(wdot)
  # wdot = applySavitzkyGolayFilter(wdot, window_length=101, poly_order=4)
  sampled_data.update(computeSpline("wdot", wdot, t_imu, sampling_steps, "xyz"))

  ## nominal model
  vdot_nom, wdot_nom = nominalModel(sampled_data, thrust_coeff, torque_coeff, inertia, mass, arm_length)
  sampled_data.update(computeSpline("vdot_nom", vdot_nom, sampled_data["t"], sampling_steps, "xyz"))
  sampled_data.update(computeSpline("wdot_nom", wdot_nom, sampled_data["t"], sampling_steps, "xyz"))

  # shift filtered data
  shiftFilteredSpline(sampled_data, step=shift_step)

  # shift time so that it starts from 0.0
  sampled_data["t"] -= sampled_data["t"][0]

  return sampled_data

def extract_hdf_from_bag(bag_path,save_path):
    processed_data = processBag(bag_path)
    #remove temporary directory with generated csv files
    shutil.rmtree(Path(bag_path).with_suffix(""))

    df = pd.concat({k: pd.Series(v) for k, v in processed_data.items()}, axis=1)

    file_name = Path(bag_path).stem
    with h5py.File(Path(save_path) / (file_name + ".hdf5"), "w") as f:
        for col in df.columns:
            f.create_dataset(col, data=df[col].values, dtype='f4')

# %% ../../nbs/datasets/quadrotor_pi.ipynb 5
def dl_quad_pi(
        save_path: Path, # directory the files are written to, created if it does not exist
        force_download: bool = False, # force download the dataset
        remove_download: bool = False # remove downloaded zip/extracted bags afterwards
) -> None:
    save_path = Path(save_path)
    save_path.mkdir(parents=True, exist_ok=True)

    download_dir = Path(get_tmp_benchmark_directory()) / 'Quadrotor_pi/'
    download_dir.mkdir(parents=True, exist_ok=True)
    zip_target_path = download_dir / 'bags.zip'
    url = 'https://drive.google.com/file/d/1b1PFSBlKTdrlTIurYNpTJWWEx1KIJzuR/view?usp=sharing'

    if force_download and zip_target_path.is_file():
        os.remove(zip_target_path) # Remove existing file to force re-download

    gdown.cached_download(
        url, str(zip_target_path), postprocess=gdown.extractall, fuzzy=True
    )

    for bag_path_str in glob.glob(str(download_dir / "*.bag")):
        bag_path = Path(bag_path_str)
        hdf_dir = save_path / get_parent_dir(bag_path.stem)
        hdf_dir.mkdir(parents=True, exist_ok=True)
        extract_hdf_from_bag(bag_path, hdf_dir)

    if remove_download:
        shutil.rmtree(download_dir)
   

# %% ../../nbs/datasets/quadrotor_pi.ipynb 7
quad_pi_u = ['u_0','u_1','u_2','u_3']

quad_pi_x_v = ['v_x','v_y','v_z']
quad_pi_x_q = ['q_w','q_x','q_y','q_z']
quad_pi_x_w = ['w_x','w_y','w_z']
quad_pi_x = quad_pi_x_v + quad_pi_x_q + quad_pi_x_w

quad_pi_y_vdot = ['vdot_x','vdot_y','vdot_z']
quad_pi_y_wdot = ['wdot_x','wdot_y','wdot_z']
quad_pi_y = quad_pi_y_vdot + quad_pi_y_wdot

# %% ../../nbs/datasets/quadrotor_pi.ipynb 8
BenchmarkQuadPi_Simulation = idb.BenchmarkSpecSimulation(
    name='BenchmarkQuadPi_Simulation', dataset_id='quad_pi',
    u_cols=quad_pi_u, y_cols=quad_pi_y, metric_func=identibench.metrics.rmse, 
    download_func=dl_quad_pi,
    init_window=100
)
BenchmarkQuadPi_Prediction = idb.BenchmarkSpecPrediction(
    name='BenchmarkQuadPi_Prediction', dataset_id='quad_pi',
    u_cols=quad_pi_u, y_cols=quad_pi_y, metric_func=identibench.metrics.rmse, 
    download_func=dl_quad_pi,
    init_window=100, pred_horizon=100,pred_step=100
)



"""Fill in a module description here"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/datasets/workshop.ipynb.

# %% auto 0
__all__ = ['BenchmarkWH_Simulation', 'BenchmarkWH_Prediction', 'BenchmarkSilverbox_Simulation', 'BenchmarkSilverbox_Prediction',
           'BenchmarkCascadedTanks_Simulation', 'BenchmarkCascadedTanks_Prediction', 'BenchmarkEMPS_Simulation',
           'BenchmarkEMPS_Prediction', 'BenchmarkNoisyWH_Simulation', 'BenchmarkNoisyWH_Prediction',
           'BenchmarkCED_Simulation', 'BenchmarkCED_Prediction', 'dl_wiener_hammerstein', 'dl_silverbox',
           'dl_cascaded_tanks', 'dl_emps', 'dl_noisy_wh', 'dl_ced']

# %% ../../nbs/datasets/workshop.ipynb 2
from ..utils import *
import identibench.benchmark as idb
import identibench.metrics
import nonlinear_benchmarks
import numpy as np
from nonlinear_benchmarks.utilities import Input_output_data
from pathlib import Path
import shutil

# %% ../../nbs/datasets/workshop.ipynb 7
def dl_wiener_hammerstein(
        save_path: Path, #directory the files are written to, created if it does not exist
        force_download: bool = False, # force download the dataset
        save_train_valid: bool = True, # save unsplitted train and valid datasets in 'train_valid' subdirectory
        split_idx: int = 80_000 # split index for train and valid datasets
) -> None:
    train_val, test = nonlinear_benchmarks.WienerHammerBenchMark(force_download=force_download)
    train = train_val[:split_idx]
    valid = train_val[split_idx:]

    dataset_to_hdf5(train,valid,test,save_path,train_valid=(train_val if save_train_valid else None))

# %% ../../nbs/datasets/workshop.ipynb 9
def rmse_mV(y_true: np.ndarray, y_pred: np.ndarray) -> float:
    return identibench.metrics.rmse(y_true,y_pred)*1000

# %% ../../nbs/datasets/workshop.ipynb 10
BenchmarkWH_Simulation = idb.BenchmarkSpecSimulation(
    name='BenchmarkWH_Simulation', dataset_id='wh',
    u_cols=['u0'], y_cols=['y0'], 
    metric_func=rmse_mV, 
    download_func=dl_wiener_hammerstein,
    init_window=50
)
BenchmarkWH_Prediction = idb.BenchmarkSpecPrediction(
    name='BenchmarkWH_Prediction', dataset_id='wh',
    u_cols=['u0'], y_cols=['y0'], 
    metric_func=rmse_mV, 
    download_func=dl_wiener_hammerstein,
    init_window=50, pred_horizon=100, pred_step=100
)

# %% ../../nbs/datasets/workshop.ipynb 15
def dl_silverbox(
        save_path: Path, #directory the files are written to, created if it does not exist
        force_download: bool = False, # force download the dataset
        save_train_valid: bool = True, # save unsplitted train and valid datasets in 'train_valid' subdirectory
        split_idx: int = 50_000 # split index for train and valid datasets
) -> None:
    train_val, test = nonlinear_benchmarks.Silverbox(force_download=force_download)
    train = train_val[:split_idx]
    valid = train_val[split_idx:]

    dataset_to_hdf5(train,valid,test,save_path,train_valid=(train_val if save_train_valid else None))

# %% ../../nbs/datasets/workshop.ipynb 17
def evaluate_silverbox(results: list[tuple[np.ndarray, np.ndarray]], spec: idb.BenchmarkSpecBase) -> dict[str, float]:

    test_configs = [
        ('test_0.hdf5', 'multisine_rmse'),
        ('test_1.hdf5', 'arrow_full_rmse'),
        ('test_2.hdf5', 'arrow_no_extrapolation_rmse'),
    ]

    aggregated_scores = {}
    for filename_part, score_name in test_configs:
        # Find the index for the current filename part
        idx = next(i for i, s in enumerate(spec.test_files) if filename_part in s.name) 
        aggregated_scores.update(
            idb.aggregate_metric_score([results[idx]], spec.metric_func, score_name=score_name)
        )

    return aggregated_scores

# %% ../../nbs/datasets/workshop.ipynb 18
BenchmarkSilverbox_Simulation = idb.BenchmarkSpecSimulation(
    name='BenchmarkSilverbox_Simulation', dataset_id='silverbox',
    u_cols=['u0'], y_cols=['y0'], 
    metric_func=rmse_mV, 
    download_func=dl_silverbox,
    custom_test_evaluation=evaluate_silverbox,
    init_window=50
)
BenchmarkSilverbox_Prediction = idb.BenchmarkSpecPrediction(
    name='BenchmarkSilverbox_Prediction', dataset_id='silverbox',
    u_cols=['u0'], y_cols=['y0'], 
    metric_func=rmse_mV, 
    download_func=dl_silverbox,
    custom_test_evaluation=evaluate_silverbox,
    init_window=50, pred_horizon=100, pred_step=100
)

# %% ../../nbs/datasets/workshop.ipynb 23
def dl_cascaded_tanks(
        save_path: Path, #directory the files are written to, created if it does not exist
        force_download: bool = False, # force download the dataset
        save_train_valid: bool = True, # save unsplitted train and valid datasets in 'train_valid' subdirectory
        split_idx: int = 160 # split index for train and valid datasets
) -> None:
    train_val, test = nonlinear_benchmarks.Cascaded_Tanks(force_download=force_download)
    train = train_val[split_idx:]
    valid = train_val[:split_idx]

    dataset_to_hdf5(train,valid,test,save_path,train_valid=(train_val if save_train_valid else None))

# %% ../../nbs/datasets/workshop.ipynb 25
BenchmarkCascadedTanks_Simulation = idb.BenchmarkSpecSimulation(
    name='BenchmarkCascadedTanks_Simulation', dataset_id='cascaded_tanks',
    u_cols=['u0'], y_cols=['y0'], metric_func=identibench.metrics.rmse, 
    download_func=dl_cascaded_tanks,
    init_window=50
)
BenchmarkCascadedTanks_Prediction = idb.BenchmarkSpecPrediction(
    name='BenchmarkCascadedTanks_Prediction', dataset_id='cascaded_tanks',
    u_cols=['u0'], y_cols=['y0'], metric_func=identibench.metrics.rmse, 
    download_func=dl_cascaded_tanks,
    init_window=50, pred_horizon=100, pred_step=100
)

# %% ../../nbs/datasets/workshop.ipynb 30
def dl_emps(
        save_path: Path, #directory the files are written to, created if it does not exist
        force_download: bool = False, # force download the dataset
        save_train_valid: bool = True, # save unsplitted train and valid datasets in 'train_valid' subdirectory
        split_idx: int = 18_000 # split index for train and valid datasets
) -> None:
    train_val, test = nonlinear_benchmarks.EMPS(force_download=force_download)
    train = train_val[:split_idx]
    valid = train_val[split_idx:]

    dataset_to_hdf5(train,valid,test,save_path,train_valid=(train_val if save_train_valid else None))

# %% ../../nbs/datasets/workshop.ipynb 32
BenchmarkEMPS_Simulation = idb.BenchmarkSpecSimulation(
    name='BenchmarkEMPS_Simulation', dataset_id='emps',
    u_cols=['u0'], y_cols=['y0'], metric_func=rmse_mV, 
    download_func=dl_emps,
    init_window=20
)
BenchmarkEMPS_Prediction = idb.BenchmarkSpecPrediction(
    name='BenchmarkEMPS_Prediction', dataset_id='emps',
    u_cols=['u0'], y_cols=['y0'], metric_func=rmse_mV, 
    download_func=dl_emps,
    init_window=20, pred_horizon=500, pred_step=100
)

# %% ../../nbs/datasets/workshop.ipynb 36
from scipy.io import loadmat
def dl_noisy_wh(
        save_path: Path, #directory the files are written to, created if it does not exist
        force_download: bool = False, # force download the dataset
        save_train_valid: bool = True # save unsplitted train and valid datasets in 'train_valid' subdirectory
) -> None:
    'the wiener hammerstein dataset with process noise'

    #extract raw .mat files, to preserve filenames necessary for train, valid split
    matfiles = nonlinear_benchmarks.not_splitted_benchmarks.WienerHammerstein_Process_Noise(data_file_locations=True,train_test_split=False,force_download=force_download)

    for file in matfiles:
        f_path = Path(file)
        save_path = Path(save_path)

        if 'Test' in f_path.stem:
            hdf_path = save_path / 'test'
        elif 'Combined' in f_path.stem:
            hdf_path = save_path / 'valid'
        else:
            hdf_path = save_path / 'train'

        out = loadmat(f_path)
        _,u,y,fs = out['dataMeas'][0,0]
        fs = fs[0,0]
        for idx,(ui,yi) in enumerate(zip(u.T,y.T)):
            iodata = Input_output_data(u=ui,y=yi, sampling_time=1/fs)
            fname = f'{f_path.stem}_{idx+1}'
            iodata_to_hdf5(iodata,hdf_path,fname)
    if save_train_valid:
        #copy train and valid files to train_valid directory
        for d in ['train','valid']:
            for f in (Path(save_path)/d).glob('*.hdf5') :
                shutil.copy2(f, (p:=Path(save_path)/'train_valid').mkdir(exist_ok=True) or p)


# %% ../../nbs/datasets/workshop.ipynb 38
BenchmarkNoisyWH_Simulation = idb.BenchmarkSpecSimulation(
    name='BenchmarkNoisyWH_Simulation', dataset_id='noisy_wh',
    u_cols=['u0'], y_cols=['y0'], metric_func=rmse_mV, 
    download_func=dl_noisy_wh,
    init_window=100
)
BenchmarkNoisyWH_Prediction = idb.BenchmarkSpecPrediction(
    name='BenchmarkNoisyWH_Prediction', dataset_id='noisy_wh',
    u_cols=['u0'], y_cols=['y0'], metric_func=rmse_mV, 
    download_func=dl_noisy_wh,
    init_window=100, pred_horizon=100, pred_step=100
)

# %% ../../nbs/datasets/workshop.ipynb 47
def dl_ced(
        save_path: Path, #directory the files are written to, created if it does not exist
        force_download: bool = False, # force download the dataset
        save_train_valid: bool = True, # save unsplitted train and valid datasets in 'train_valid' subdirectory
        split_idx: int = 300 # split index for train and valid datasets
) -> None:
    train_val, test = nonlinear_benchmarks.CED(force_download=force_download,always_return_tuples_of_datasets=True)
    train = tuple(x[:split_idx] for x in train_val)
    valid = tuple(x[split_idx:] for x in train_val)

    dataset_to_hdf5(train,valid,test,save_path,train_valid=(train_val if save_train_valid else None))

# %% ../../nbs/datasets/workshop.ipynb 49
def evaluate_ced(results: list[tuple[np.ndarray, np.ndarray]], spec: idb.BenchmarkSpecBase) -> dict[str, float]:
    test_configs = [
        ('test_0.hdf5', 'test_1_rmse'),
        ('test_1.hdf5', 'test_2_rmse'),
    ]

    aggregated_scores = {}
    for filename_part, score_name in test_configs:
        # Find the index for the current filename part
        idx = next(i for i, s in enumerate(spec.test_files) if filename_part in s.name) 
        aggregated_scores.update(
            idb.aggregate_metric_score([results[idx]], spec.metric_func, score_name=score_name)
        )

    return aggregated_scores

# %% ../../nbs/datasets/workshop.ipynb 50
BenchmarkCED_Simulation = idb.BenchmarkSpecSimulation(
    name='BenchmarkCED_Simulation', dataset_id='ced',
    u_cols=['u0'], y_cols=['y0'], metric_func=identibench.metrics.rmse, 
    download_func=dl_ced,
    custom_test_evaluation=evaluate_ced,
    init_window=10
)
BenchmarkCED_Prediction = idb.BenchmarkSpecPrediction(
    name='BenchmarkCED_Prediction', dataset_id='ced',
    u_cols=['u0'], y_cols=['y0'], metric_func=identibench.metrics.rmse, 
    download_func=dl_ced,
    custom_test_evaluation=evaluate_ced,
    init_window=10, pred_horizon=30, pred_step=30
)



"""Fill in a module description here"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/datasets/ship.ipynb.

# %% auto 0
__all__ = ['ship_u', 'ship_y', 'BenchmarkShip_Simulation', 'BenchmarkShip_Prediction', 'dl_ship']

# %% ../../nbs/datasets/ship.ipynb 2
from nonlinear_benchmarks.utilities import get_tmp_benchmark_directory
import identibench.benchmark as idb
import identibench.metrics
from pathlib import Path
import os
import h5py
import numpy as np
from easyDataverse import Dataverse
import pandas as pd
import shutil

# %% ../../nbs/datasets/ship.ipynb 3
def dl_ship(
        save_path: Path, #directory the files are written to, created if it does not exist
        force_download: bool = False, # force download the dataset
        remove_download: bool = True
) -> None:
    save_path = Path(save_path)
    download_dir = Path(get_tmp_benchmark_directory()) / 'Ship'

    if force_download and download_dir.exists():
        print(f"Force reload: Removing existing directory: {download_dir}")
        shutil.rmtree(download_dir)

    dataverse = Dataverse('https://darus.uni-stuttgart.de/')
    dataverse.load_dataset(
        pid='doi:10.18419/darus-2905',
        filedir=download_dir,
    )

    #str to Path to be plattform independent
    structure_mapping = {
        Path('patrol_ship_routine/processed/train'): 'train',
        Path('patrol_ship_routine/processed/validation'): 'valid',
        Path('patrol_ship_routine/processed/test'): 'test',
        Path('patrol_ship_ood/processed/test'): 'test_ood'
    }

    # Ensure desired directories exist
    for subdir in structure_mapping.values():
        os.makedirs(os.path.join(save_path, subdir), exist_ok=True)

    def convert_tab_to_hdf5(tab_path: Path, hdf5_path: Path) -> None:
        df = pd.read_csv(tab_path, sep='\t')
        with h5py.File(hdf5_path, 'w') as hdf:
            for column in df.columns:
                data = df[column].astype(np.float32).values
                hdf.create_dataset(column, data=data, dtype='f4')

    # Walk through the current directory structure and process files
    for subdir, dirs, files in os.walk(download_dir):
        for file in files:
            if file.endswith('.tab'):
                current_file_path = os.path.join(subdir, file)
                
                # Determine the relative path
                relative_subdir = Path(os.path.relpath(subdir, download_dir))
                
                # Find the corresponding desired subdir
                if relative_subdir in structure_mapping:
                    desired_subdir = structure_mapping[relative_subdir]
                    
                    # Construct desired file paths
                    base_filename = file.replace('.tab', '')
                    desired_hdf5_path = os.path.join(save_path, desired_subdir, base_filename + '.hdf5')
                    
                    convert_tab_to_hdf5(current_file_path, desired_hdf5_path)

    #remove downloaded files
    if remove_download:
        shutil.rmtree(download_dir)

# %% ../../nbs/datasets/ship.ipynb 5
ship_u = ['n','deltal','deltar','Vw']
ship_y = ['alpha_x','alpha_y','u','v','p','r','phi']

BenchmarkShip_Simulation = idb.BenchmarkSpecSimulation(
    name='BenchmarkShip_Simulation', dataset_id='ship',
    u_cols=ship_u, y_cols=ship_y, metric_func=identibench.metrics.rmse, 
    download_func=dl_ship,
    init_window=100
)
BenchmarkShip_Prediction = idb.BenchmarkSpecPrediction(
    name='BenchmarkShip_Prediction', dataset_id='ship',
    u_cols=ship_u, y_cols=ship_y, metric_func=identibench.metrics.rmse, 
    download_func=dl_ship,
    init_window=100, pred_horizon=100,pred_step=100
)



"""Fill in a module description here"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/datasets/quad_pelican.ipynb.

# %% auto 0
__all__ = ['pelican_u_motors', 'pelican_u_motors_cmd', 'pelican_y_euler', 'pelican_y_euler_rates', 'pelican_y_pos',
           'pelican_y_vel', 'pelican_y_rate', 'pelican_u', 'pelican_y', 'BenchmarkQuadPelican_Simulation',
           'BenchmarkQuadPelican_Prediction', 'dl_quad_pelican']

# %% ../../nbs/datasets/quad_pelican.ipynb 2
from nonlinear_benchmarks.utilities import cashed_download
import identibench.benchmark as idb
import identibench.metrics
from pathlib import Path
import os
import h5py
import numpy as np
import scipy.io as sio

# %% ../../nbs/datasets/quad_pelican.ipynb 3
pelican_fnames_train = ['hdf5flight24.hdf5',
                        'hdf5flight25.hdf5',
                        'hdf5flight38.hdf5',
                        'hdf5flight20.hdf5',
                        'hdf5flight26.hdf5',
                        'hdf5flight14.hdf5',
                        'hdf5flight21.hdf5',
                        'hdf5flight11.hdf5',
                        'hdf5flight40.hdf5',
                        'hdf5flight9.hdf5',
                        'hdf5flight23.hdf5',
                        'hdf5flight19.hdf5',
                        'hdf5flight27.hdf5',
                        'hdf5flight12.hdf5',
                        'hdf5flight6.hdf5',
                        'hdf5flight50.hdf5',
                        'hdf5flight36.hdf5',
                        'hdf5flight48.hdf5',
                        'hdf5flight28.hdf5',
                        'hdf5flight44.hdf5',
                        'hdf5flight34.hdf5',
                        'hdf5flight32.hdf5',
                        'hdf5flight3.hdf5',
                        'hdf5flight45.hdf5',
                        'hdf5flight33.hdf5',
                        'hdf5flight4.hdf5']

pelican_fnames_valid =[  'hdf5flight10.hdf5',
                         'hdf5flight15.hdf5',
                         'hdf5flight2.hdf5',
                         'hdf5flight18.hdf5',
                         'hdf5flight51.hdf5',
                         'hdf5flight52.hdf5',
                         'hdf5flight35.hdf5',
                         'hdf5flight13.hdf5',
                         'hdf5flight22.hdf5',
                         'hdf5flight53.hdf5']

pelican_fnames_test = [  'hdf5flight8.hdf5',
                         'hdf5flight16.hdf5',
                         'hdf5flight5.hdf5',
                         'hdf5flight7.hdf5',
                         'hdf5flight41.hdf5',
                         'hdf5flight1.hdf5',
                         'hdf5flight17.hdf5',
                         'hdf5flight37.hdf5',
                         'hdf5flight30.hdf5',
                         'hdf5flight49.hdf5',
                         'hdf5flight29.hdf5',
                         'hdf5flight31.hdf5',
                         'hdf5flight39.hdf5',
                         'hdf5flight54.hdf5',
                         'hdf5flight47.hdf5',
                         'hdf5flight43.hdf5',
                         'hdf5flight42.hdf5',
                         'hdf5flight46.hdf5']

def get_parent_dir(f_name: str # name of the flight
                  ) -> str:
    if f_name in pelican_fnames_train:
        return 'train'
    elif f_name in pelican_fnames_valid:
        return 'valid'
    elif f_name in pelican_fnames_test:
        return 'test'
    else:
        return ValueError(f'Filename {f_name} not recognized!')

# %% ../../nbs/datasets/quad_pelican.ipynb 4
def dl_quad_pelican(
        save_path: Path, #directory the files are written to, created if it does not exist
        force_download: bool = False, # force download the dataset
        remove_download: bool = False
) -> None:
    save_path = Path(save_path)
    url_pelican = 'http://wavelab.uwaterloo.ca/wp-content/uploads/2017/09/AscTec_Pelican_Flight_Dataset.mat'
   
    tmp_dir = cashed_download(url_pelican,'Quad_pelican',zipped=False,force_download=force_download)
    downloaded_fname = Path(tmp_dir) / Path(url_pelican).name
    
    def write_signal(fname: Path, sname: str, signal: np.ndarray) -> None:
        with h5py.File(fname, 'a') as f:
            for i in range(signal.shape[1]):
                ds_name = f'{sname}{i+1}'
                sig = signal[:, i]
                f.create_dataset(ds_name, data=sig, dtype='f4')
    
    flight_data = sio.loadmat(downloaded_fname,simplify_cells=True)
    flights = flight_data['flights']
    
    for k, flight in enumerate(flights, start=1):
        f_name = f'hdf5flight{k}.hdf5'
        parent_dir = Path(save_path) / get_parent_dir(f_name)
        f_path = parent_dir / f_name

        os.makedirs(parent_dir, exist_ok=True)
        
        if os.path.exists(f_path):
            os.remove(f_path)

        write_signal(f_path, 'vel', flight['Vel'])
        write_signal(f_path, 'pos', flight['Pos'][1:, :])
        write_signal(f_path, 'euler', flight['Euler'][1:, :])
        write_signal(f_path, 'euler_rates', flight['Euler_Rates'])
        write_signal(f_path, 'motors', flight['Motors'][1:, :])
        write_signal(f_path, 'motors_cmd', flight['Motors_CMD'][1:, :])
        write_signal(f_path, 'pqr', flight['pqr'][:-1, :])
    
    # %%
    #cleanup downloaded quadrotor file
    if remove_download: os.remove(downloaded_fname)

# %% ../../nbs/datasets/quad_pelican.ipynb 6
pelican_u_motors = [f'motors{i}' for i in range(1,4+1)]
pelican_u_motors_cmd = [f'motors_cmd{i}' for i in range(1,4+1)]

pelican_y_euler = [f'euler{i}' for i in range(1,3+1)]
pelican_y_euler_rates = [f'euler_rates{i}' for i in range(1,3+1)]
pelican_y_pos = [f'pos{i}' for i in range(1,3+1)]
pelican_y_vel = [f'vel{i}' for i in range(1,3+1)]
pelican_y_rate = [f'pqr{i}' for i in range(1,3+1)]

pelican_u = pelican_u_motors
pelican_y = pelican_y_euler_rates+pelican_y_vel

# %% ../../nbs/datasets/quad_pelican.ipynb 7
BenchmarkQuadPelican_Simulation = idb.BenchmarkSpecSimulation(
    name='BenchmarkQuadPelican_Simulation', dataset_id='quad_pelican',
    u_cols=pelican_u, y_cols=pelican_y, metric_func=identibench.metrics.rmse, 
    download_func=dl_quad_pelican,
    init_window=100
)
BenchmarkQuadPelican_Prediction = idb.BenchmarkSpecPrediction(
    name='BenchmarkQuadPelican_Prediction', dataset_id='quad_pelican',
    u_cols=pelican_u, y_cols=pelican_y, metric_func=identibench.metrics.rmse, 
    download_func=dl_quad_pelican,
    init_window=100, pred_horizon=100,pred_step=100
)



from pathlib import Path
from .workshop import dl_wiener_hammerstein, dl_silverbox, dl_cascaded_tanks, dl_emps, dl_noisy_wh,dl_ced
from .industrial_robot import dl_robot_forward, dl_robot_inverse
from .ship import dl_ship
from .quad_pelican import dl_quad_pelican
from .quad_pi import dl_quad_pi
from .broad import dl_broad

all_dataset_loaders = {
    'wiener_hammerstein':  dl_wiener_hammerstein, 'silverbox': dl_silverbox, 'cascaded_tanks': dl_cascaded_tanks,
    'emps': dl_emps, 'noisy_wh': dl_noisy_wh, 'robot_forward': dl_robot_forward, 'robot_inverse': dl_robot_inverse,
    'ship': dl_ship, 'quad_pelican': dl_quad_pelican, 'quad_pi': dl_quad_pi, 'broad': dl_broad,
    'ced': dl_ced
}

def download_all_datasets(save_path: Path, force_download: bool = False):
    """Download all datasets provided by identibench.datasets into subdirectories."""
    save_path = Path(save_path)
    print(f"Downloading all datasets to {save_path}...")
    for name, loader in all_dataset_loaders.items():
        print(f"--- Downloading/Preparing {name} ---")
        try:
            loader(save_path / name, force_download=force_download)
        except Exception as e:
            print(f"ERROR downloading {name}: {e}")
    print("--- Finished downloading all datasets ---")


__all__ = ['all_dataset_loaders', 'download_all_datasets']


"""Fill in a module description here"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/datasets/broad.ipynb.

# %% auto 0
__all__ = ['dl_broad']

# %% ../../nbs/datasets/broad.ipynb 2
from ..utils import write_array
from pathlib import Path
import os
import h5py
import requests
from io import BytesIO

# %% ../../nbs/datasets/broad.ipynb 3
def dl_broad(
        save_path: Path, #directory the files are written to, created if it does not exist
        force_download: bool = True, # force download the dataset
) -> None:
    save_path = Path(save_path)

    idxs_valid = ['14_', '39_', '21_']
    idxs_test = ['29_', '22_', '35_']
    
    url = "https://api.github.com/repos/dlaidig/broad/contents/data_hdf5"
    response = requests.get(url)
    data = response.json()

    for file in data:
        # Check if it is a file in hdf5 format
        if (file['type'] == 'file') and ('.hdf5' in file['name']):
            download_url = file['download_url']
            file_response = requests.get(download_url)
            file_response.raise_for_status()

            file_idx = file['name'][:3]
            if file_idx in idxs_valid:
                parent = 'valid'
            elif file_idx in idxs_test:
                parent = 'test'
            else:
                parent = 'train'
            hdf_path = save_path / parent 
            os.makedirs(hdf_path, exist_ok=True)

            #open loaded hdf5 file in ram
            tmp_file = BytesIO(file_response.content)
            # Write file to local system
            with h5py.File(tmp_file) as f_read:
                with h5py.File(hdf_path / file['name'],'w') as f_write:
                    def transfer_dataset(ds_name: str) -> None:
                        x = f_read[ds_name][:]
                        if x.ndim == 2:
                            write_array(f_write,ds_name,x,dtype='f4')

                    #transfer each dataset in the source file
                    f_read.visit(transfer_dataset)



__version__ = "0.1.4"

from .benchmark import run_benchmark,run_multiple_benchmarks, BenchmarkSpecSimulation, BenchmarkSpecPrediction, TrainingContext
from . import metrics
from . import datasets

# Workshop Benchmarks
from .datasets.workshop import (
    BenchmarkWH_Simulation, BenchmarkWH_Prediction,
    BenchmarkSilverbox_Simulation, BenchmarkSilverbox_Prediction,
    BenchmarkCascadedTanks_Simulation, BenchmarkCascadedTanks_Prediction,
    BenchmarkEMPS_Simulation, BenchmarkEMPS_Prediction,
    BenchmarkNoisyWH_Simulation, BenchmarkNoisyWH_Prediction,
    BenchmarkCED_Simulation, BenchmarkCED_Prediction
)
# Robot Benchmarks
from .datasets.industrial_robot import (
    BenchmarkRobotForward_Simulation, BenchmarkRobotForward_Prediction,
    BenchmarkRobotInverse_Simulation, BenchmarkRobotInverse_Prediction
)
# Ship Benchmark
from .datasets.ship import BenchmarkShip_Simulation, BenchmarkShip_Prediction
# Quadrotor Benchmarks
from .datasets.quad_pelican import BenchmarkQuadPelican_Simulation, BenchmarkQuadPelican_Prediction
from .datasets.quad_pi import BenchmarkQuadPi_Simulation, BenchmarkQuadPi_Prediction
# Broad Benchmark
# from .datasets.broad import BenchmarkBroad_Simulation, BenchmarkBroad_Prediction

simulation_benchmarks = {
    'WH_Sim': BenchmarkWH_Simulation,
    'Silverbox_Sim': BenchmarkSilverbox_Simulation,
    'Tanks_Sim': BenchmarkCascadedTanks_Simulation,
    'EMPS_Sim': BenchmarkEMPS_Simulation,
    'NoisyWH_Sim': BenchmarkNoisyWH_Simulation,
    'RobotForward_Sim': BenchmarkRobotForward_Simulation,
    'RobotInverse_Sim': BenchmarkRobotInverse_Simulation,
    'Ship_Sim': BenchmarkShip_Simulation,
    'QuadPelican_Sim': BenchmarkQuadPelican_Simulation,
    'QuadPi_Sim': BenchmarkQuadPi_Simulation,
}

prediction_benchmarks = {
    'WH_Pred': BenchmarkWH_Prediction,
    'Silverbox_Pred': BenchmarkSilverbox_Prediction,
    'Tanks_Pred': BenchmarkCascadedTanks_Prediction,
    'EMPS_Pred': BenchmarkEMPS_Prediction,
    'NoisyWH_Pred': BenchmarkNoisyWH_Prediction,
    'RobotForward_Pred': BenchmarkRobotForward_Prediction,
    'RobotInverse_Pred': BenchmarkRobotInverse_Prediction,
    'Ship_Pred': BenchmarkShip_Prediction,
    'QuadPelican_Pred': BenchmarkQuadPelican_Prediction,
    'QuadPi_Pred': BenchmarkQuadPi_Prediction,
}

all_benchmarks = {**simulation_benchmarks, **prediction_benchmarks}

workshop_benchmarks = {
    'WH_Sim': BenchmarkWH_Simulation,
    'Silverbox_Sim': BenchmarkSilverbox_Simulation,
    'EMPS_Sim': BenchmarkEMPS_Simulation,
    'CED_Sim': BenchmarkCED_Simulation,
}


# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/identibench',
                'doc_host': 'https://daniel-om-weber.github.io',
                'git_url': 'https://github.com/daniel-om-weber/identibench',
                'lib_path': 'identibench'},
  'syms': { 'identibench.benchmark': { 'identibench.benchmark.BenchmarkSpecBase': ( 'benchmark.html#benchmarkspecbase',
                                                                                    'identibench/benchmark.py'),
                                       'identibench.benchmark.BenchmarkSpecBase.__init__': ( 'benchmark.html#benchmarkspecbase.__init__',
                                                                                             'identibench/benchmark.py'),
                                       'identibench.benchmark.BenchmarkSpecBase.data_root': ( 'benchmark.html#benchmarkspecbase.data_root',
                                                                                              'identibench/benchmark.py'),
                                       'identibench.benchmark.BenchmarkSpecBase.dataset_path': ( 'benchmark.html#benchmarkspecbase.dataset_path',
                                                                                                 'identibench/benchmark.py'),
                                       'identibench.benchmark.BenchmarkSpecBase.ensure_dataset_exists': ( 'benchmark.html#benchmarkspecbase.ensure_dataset_exists',
                                                                                                          'identibench/benchmark.py'),
                                       'identibench.benchmark.BenchmarkSpecBase.test_files': ( 'benchmark.html#benchmarkspecbase.test_files',
                                                                                               'identibench/benchmark.py'),
                                       'identibench.benchmark.BenchmarkSpecBase.test_path': ( 'benchmark.html#benchmarkspecbase.test_path',
                                                                                              'identibench/benchmark.py'),
                                       'identibench.benchmark.BenchmarkSpecBase.train_files': ( 'benchmark.html#benchmarkspecbase.train_files',
                                                                                                'identibench/benchmark.py'),
                                       'identibench.benchmark.BenchmarkSpecBase.train_path': ( 'benchmark.html#benchmarkspecbase.train_path',
                                                                                               'identibench/benchmark.py'),
                                       'identibench.benchmark.BenchmarkSpecBase.train_valid_files': ( 'benchmark.html#benchmarkspecbase.train_valid_files',
                                                                                                      'identibench/benchmark.py'),
                                       'identibench.benchmark.BenchmarkSpecBase.train_valid_path': ( 'benchmark.html#benchmarkspecbase.train_valid_path',
                                                                                                     'identibench/benchmark.py'),
                                       'identibench.benchmark.BenchmarkSpecBase.valid_files': ( 'benchmark.html#benchmarkspecbase.valid_files',
                                                                                                'identibench/benchmark.py'),
                                       'identibench.benchmark.BenchmarkSpecBase.valid_path': ( 'benchmark.html#benchmarkspecbase.valid_path',
                                                                                               'identibench/benchmark.py'),
                                       'identibench.benchmark.BenchmarkSpecPrediction': ( 'benchmark.html#benchmarkspecprediction',
                                                                                          'identibench/benchmark.py'),
                                       'identibench.benchmark.BenchmarkSpecPrediction.__init__': ( 'benchmark.html#benchmarkspecprediction.__init__',
                                                                                                   'identibench/benchmark.py'),
                                       'identibench.benchmark.BenchmarkSpecSimulation': ( 'benchmark.html#benchmarkspecsimulation',
                                                                                          'identibench/benchmark.py'),
                                       'identibench.benchmark.BenchmarkSpecSimulation.__init__': ( 'benchmark.html#benchmarkspecsimulation.__init__',
                                                                                                   'identibench/benchmark.py'),
                                       'identibench.benchmark.TrainingContext': ( 'benchmark.html#trainingcontext',
                                                                                  'identibench/benchmark.py'),
                                       'identibench.benchmark.TrainingContext.__init__': ( 'benchmark.html#trainingcontext.__init__',
                                                                                           'identibench/benchmark.py'),
                                       'identibench.benchmark.TrainingContext.get_train_sequences': ( 'benchmark.html#trainingcontext.get_train_sequences',
                                                                                                      'identibench/benchmark.py'),
                                       'identibench.benchmark.TrainingContext.get_train_valid_sequences': ( 'benchmark.html#trainingcontext.get_train_valid_sequences',
                                                                                                            'identibench/benchmark.py'),
                                       'identibench.benchmark.TrainingContext.get_valid_sequences': ( 'benchmark.html#trainingcontext.get_valid_sequences',
                                                                                                      'identibench/benchmark.py'),
                                       'identibench.benchmark._dummy_build_model': ( 'benchmark.html#_dummy_build_model',
                                                                                     'identibench/benchmark.py'),
                                       'identibench.benchmark._test_prediction': ( 'benchmark.html#_test_prediction',
                                                                                   'identibench/benchmark.py'),
                                       'identibench.benchmark._test_simulation': ( 'benchmark.html#_test_simulation',
                                                                                   'identibench/benchmark.py'),
                                       'identibench.benchmark.aggregate_metric_score': ( 'benchmark.html#aggregate_metric_score',
                                                                                         'identibench/benchmark.py'),
                                       'identibench.benchmark.run_benchmark': ('benchmark.html#run_benchmark', 'identibench/benchmark.py'),
                                       'identibench.benchmark.run_multiple_benchmarks': ( 'benchmark.html#run_multiple_benchmarks',
                                                                                          'identibench/benchmark.py')},
            'identibench.datasets.broad': { 'identibench.datasets.broad.dl_broad': ( 'datasets/broad.html#dl_broad',
                                                                                     'identibench/datasets/broad.py')},
            'identibench.datasets.industrial_robot': { 'identibench.datasets.industrial_robot.dl_robot_forward': ( 'datasets/industrial_robot.html#dl_robot_forward',
                                                                                                                   'identibench/datasets/industrial_robot.py'),
                                                       'identibench.datasets.industrial_robot.dl_robot_inverse': ( 'datasets/industrial_robot.html#dl_robot_inverse',
                                                                                                                   'identibench/datasets/industrial_robot.py'),
                                                       'identibench.datasets.industrial_robot.robot_mat2hdf': ( 'datasets/industrial_robot.html#robot_mat2hdf',
                                                                                                                'identibench/datasets/industrial_robot.py')},
            'identibench.datasets.quad_pelican': { 'identibench.datasets.quad_pelican.dl_quad_pelican': ( 'datasets/quad_pelican.html#dl_quad_pelican',
                                                                                                          'identibench/datasets/quad_pelican.py'),
                                                   'identibench.datasets.quad_pelican.get_parent_dir': ( 'datasets/quad_pelican.html#get_parent_dir',
                                                                                                         'identibench/datasets/quad_pelican.py')},
            'identibench.datasets.quad_pi': { 'identibench.datasets.quad_pi.appendHistory': ( 'datasets/quadrotor_pi.html#appendhistory',
                                                                                              'identibench/datasets/quad_pi.py'),
                                              'identibench.datasets.quad_pi.applyButterLowpassFilter': ( 'datasets/quadrotor_pi.html#applybutterlowpassfilter',
                                                                                                         'identibench/datasets/quad_pi.py'),
                                              'identibench.datasets.quad_pi.applySavitzkyGolayFilter': ( 'datasets/quadrotor_pi.html#applysavitzkygolayfilter',
                                                                                                         'identibench/datasets/quad_pi.py'),
                                              'identibench.datasets.quad_pi.butterLowpassFilter': ( 'datasets/quadrotor_pi.html#butterlowpassfilter',
                                                                                                    'identibench/datasets/quad_pi.py'),
                                              'identibench.datasets.quad_pi.computeSpline': ( 'datasets/quadrotor_pi.html#computespline',
                                                                                              'identibench/datasets/quad_pi.py'),
                                              'identibench.datasets.quad_pi.differentiate': ( 'datasets/quadrotor_pi.html#differentiate',
                                                                                              'identibench/datasets/quad_pi.py'),
                                              'identibench.datasets.quad_pi.differentiateFivePointStencil': ( 'datasets/quadrotor_pi.html#differentiatefivepointstencil',
                                                                                                              'identibench/datasets/quad_pi.py'),
                                              'identibench.datasets.quad_pi.dl_quad_pi': ( 'datasets/quadrotor_pi.html#dl_quad_pi',
                                                                                           'identibench/datasets/quad_pi.py'),
                                              'identibench.datasets.quad_pi.dropNoise': ( 'datasets/quadrotor_pi.html#dropnoise',
                                                                                          'identibench/datasets/quad_pi.py'),
                                              'identibench.datasets.quad_pi.extract_hdf_from_bag': ( 'datasets/quadrotor_pi.html#extract_hdf_from_bag',
                                                                                                     'identibench/datasets/quad_pi.py'),
                                              'identibench.datasets.quad_pi.get_parent_dir': ( 'datasets/quadrotor_pi.html#get_parent_dir',
                                                                                               'identibench/datasets/quad_pi.py'),
                                              'identibench.datasets.quad_pi.nominalModel': ( 'datasets/quadrotor_pi.html#nominalmodel',
                                                                                             'identibench/datasets/quad_pi.py'),
                                              'identibench.datasets.quad_pi.parseBag': ( 'datasets/quadrotor_pi.html#parsebag',
                                                                                         'identibench/datasets/quad_pi.py'),
                                              'identibench.datasets.quad_pi.processBag': ( 'datasets/quadrotor_pi.html#processbag',
                                                                                           'identibench/datasets/quad_pi.py'),
                                              'identibench.datasets.quad_pi.savitzkyGolayFilter': ( 'datasets/quadrotor_pi.html#savitzkygolayfilter',
                                                                                                    'identibench/datasets/quad_pi.py'),
                                              'identibench.datasets.quad_pi.shift': ( 'datasets/quadrotor_pi.html#shift',
                                                                                      'identibench/datasets/quad_pi.py'),
                                              'identibench.datasets.quad_pi.shiftFilteredSpline': ( 'datasets/quadrotor_pi.html#shiftfilteredspline',
                                                                                                    'identibench/datasets/quad_pi.py')},
            'identibench.datasets.ship': { 'identibench.datasets.ship.dl_ship': ( 'datasets/ship.html#dl_ship',
                                                                                  'identibench/datasets/ship.py')},
            'identibench.datasets.workshop': { 'identibench.datasets.workshop.dl_cascaded_tanks': ( 'datasets/workshop.html#dl_cascaded_tanks',
                                                                                                    'identibench/datasets/workshop.py'),
                                               'identibench.datasets.workshop.dl_ced': ( 'datasets/workshop.html#dl_ced',
                                                                                         'identibench/datasets/workshop.py'),
                                               'identibench.datasets.workshop.dl_emps': ( 'datasets/workshop.html#dl_emps',
                                                                                          'identibench/datasets/workshop.py'),
                                               'identibench.datasets.workshop.dl_noisy_wh': ( 'datasets/workshop.html#dl_noisy_wh',
                                                                                              'identibench/datasets/workshop.py'),
                                               'identibench.datasets.workshop.dl_silverbox': ( 'datasets/workshop.html#dl_silverbox',
                                                                                               'identibench/datasets/workshop.py'),
                                               'identibench.datasets.workshop.dl_wiener_hammerstein': ( 'datasets/workshop.html#dl_wiener_hammerstein',
                                                                                                        'identibench/datasets/workshop.py'),
                                               'identibench.datasets.workshop.evaluate_ced': ( 'datasets/workshop.html#evaluate_ced',
                                                                                               'identibench/datasets/workshop.py'),
                                               'identibench.datasets.workshop.evaluate_silverbox': ( 'datasets/workshop.html#evaluate_silverbox',
                                                                                                     'identibench/datasets/workshop.py'),
                                               'identibench.datasets.workshop.rmse_mV': ( 'datasets/workshop.html#rmse_mv',
                                                                                          'identibench/datasets/workshop.py')},
            'identibench.metrics': { 'identibench.metrics.fit_index': ('metrics.html#fit_index', 'identibench/metrics.py'),
                                     'identibench.metrics.mae': ('metrics.html#mae', 'identibench/metrics.py'),
                                     'identibench.metrics.nrmse': ('metrics.html#nrmse', 'identibench/metrics.py'),
                                     'identibench.metrics.r_squared': ('metrics.html#r_squared', 'identibench/metrics.py'),
                                     'identibench.metrics.rmse': ('metrics.html#rmse', 'identibench/metrics.py')},
            'identibench.utils': { 'identibench.utils._dummy_dataset_loader': ('utils.html#_dummy_dataset_loader', 'identibench/utils.py'),
                                   'identibench.utils._load_sequences_from_files': ( 'utils.html#_load_sequences_from_files',
                                                                                     'identibench/utils.py'),
                                   'identibench.utils.dataset_to_hdf5': ('utils.html#dataset_to_hdf5', 'identibench/utils.py'),
                                   'identibench.utils.download': ('utils.html#download', 'identibench/utils.py'),
                                   'identibench.utils.get_default_data_root': ('utils.html#get_default_data_root', 'identibench/utils.py'),
                                   'identibench.utils.hdf_files_from_path': ('utils.html#hdf_files_from_path', 'identibench/utils.py'),
                                   'identibench.utils.iodata_to_hdf5': ('utils.html#iodata_to_hdf5', 'identibench/utils.py'),
                                   'identibench.utils.unrar_download': ('utils.html#unrar_download', 'identibench/utils.py'),
                                   'identibench.utils.unzip_download': ('utils.html#unzip_download', 'identibench/utils.py'),
                                   'identibench.utils.write_array': ('utils.html#write_array', 'identibench/utils.py'),
                                   'identibench.utils.write_dataset': ('utils.html#write_dataset', 'identibench/utils.py')}}}



"""Fill in a module description here"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/metrics.ipynb.

# %% auto 0
__all__ = ['rmse', 'nrmse', 'fit_index', 'mae', 'r_squared']

# %% ../nbs/metrics.ipynb 2
import numpy as np
import warnings

# %% ../nbs/metrics.ipynb 4
def rmse(
    y_true: np.ndarray, # Ground truth target values.
    y_pred: np.ndarray, # Estimated target values.
    time_axis: int = 0 # Axis representing time or samples.
    ) -> np.ndarray: # Root Mean Squared Error for each channel.
    """
    Computes the Root Mean Square Error (RMSE) along a specified time axis.

    Calculates RMSE = sqrt(mean((y_pred - y_true)**2)) separately for each channel
    defined by the remaining axes.
    """
    y_true = np.asarray(y_true)
    y_pred = np.asarray(y_pred)
    if y_true.shape != y_pred.shape:
        raise ValueError(f"Input shapes must match. Got {y_true.shape} and {y_pred.shape}")

    # Ensure time_axis is valid
    if not (0 <= time_axis < y_true.ndim):
         raise ValueError(f"Invalid time_axis {time_axis} for array with {y_true.ndim} dimensions")

    # Calculate RMSE
    try:
        rmse_val = np.sqrt(np.mean((y_pred - y_true)**2, axis=time_axis))
    except FloatingPointError as e:
         warnings.warn(f"Floating point error during RMSE calculation: {e}. Check for NaNs or Infs.", RuntimeWarning)
         # Decide how to handle - perhaps return NaN or raise a more specific error
         # For now, re-raising but ideally might return NaN or a specific value
         raise e
    return rmse_val

# %% ../nbs/metrics.ipynb 6
def nrmse(
    y_true: np.ndarray, # Ground truth target values.
    y_pred: np.ndarray, # Estimated target values.
    time_axis: int = 0, # Axis representing time or samples.
    std_tolerance: float = 1e-9 # Minimum standard deviation allowed for y_true to avoid division by zero.
    ) -> np.ndarray: # Normalized Root Mean Squared Error for each channel.
    """
    Computes the Normalized Root Mean Square Error (NRMSE).

    Calculates NRMSE = RMSE / std(y_true) separately for each channel.
    Returns NaN for channels where std(y_true) is close to zero (below std_tolerance).
    """
    rmse_val = rmse(y_true, y_pred, time_axis=time_axis)
    std_true = np.std(y_true, axis=time_axis)

    # Initialize nrmse_val with NaNs or another placeholder
    nrmse_val = np.full_like(std_true, fill_value=np.nan, dtype=np.float64)

    # Identify channels with standard deviation above the tolerance
    valid_std_mask = std_true > std_tolerance

    # Calculate NRMSE only for valid channels
    # Using np.divide with 'where' handles division by zero gracefully if needed,
    # but the explicit check is clearer here.
    if np.any(valid_std_mask):
        nrmse_val[valid_std_mask] = rmse_val[valid_std_mask] / std_true[valid_std_mask]

    # Warn if any channels had std below tolerance
    if not np.all(valid_std_mask):
        warnings.warn(f"Standard deviation of y_true is below tolerance ({std_tolerance}) for some channels. NRMSE set to NaN for these channels.", RuntimeWarning)

    return nrmse_val

# %% ../nbs/metrics.ipynb 8
def fit_index(
    y_true: np.ndarray, # Ground truth target values.
    y_pred: np.ndarray, # Estimated target values.
    time_axis: int = 0, # Axis representing time or samples.
    std_tolerance: float = 1e-9 # Minimum standard deviation allowed for y_true.
    ) -> np.ndarray: # Fit index (in percent) for each channel.
    """
    Computes the Fit Index (FIT) commonly used in System Identification.

    Calculates FIT = 100 * (1 - NRMSE) separately for each channel.
    Returns NaN for channels where NRMSE could not be calculated (e.g., std(y_true) near zero).
    """
    nrmse_val = nrmse(y_true, y_pred, time_axis=time_axis, std_tolerance=std_tolerance)

    # Fit index calculation, handles potential NaNs from nrmse
    fit_val = 100.0 * (1.0 - nrmse_val)

    return fit_val

# %% ../nbs/metrics.ipynb 10
def mae(
    y_true: np.ndarray, # Ground truth target values.
    y_pred: np.ndarray, # Estimated target values.
    time_axis: int = 0 # Axis representing time or samples.
    ) -> np.ndarray: # Mean Absolute Error for each channel.
    """
    Computes the Mean Absolute Error (MAE) along a specified time axis.

    Calculates MAE = mean(abs(y_pred - y_true)) separately for each channel
    defined by the remaining axes.
    """
    y_true = np.asarray(y_true)
    y_pred = np.asarray(y_pred)
    if y_true.shape != y_pred.shape:
        raise ValueError(f"Input shapes must match. Got {y_true.shape} and {y_pred.shape}")

    # Ensure time_axis is valid
    if not (0 <= time_axis < y_true.ndim):
         raise ValueError(f"Invalid time_axis {time_axis} for array with {y_true.ndim} dimensions")

    # Calculate MAE
    try:
        mae_val = np.mean(np.abs(y_pred - y_true), axis=time_axis)
    except FloatingPointError as e:
         warnings.warn(f"Floating point error during MAE calculation: {e}. Check for NaNs or Infs.", RuntimeWarning)
         # Decide how to handle - perhaps return NaN or raise a more specific error
         raise e
    return mae_val

# %% ../nbs/metrics.ipynb 12
def r_squared(
    y_true: np.ndarray, # Ground truth target values.
    y_pred: np.ndarray, # Estimated target values.
    time_axis: int = 0, # Axis representing time or samples.
    std_tolerance: float = 1e-9 # Minimum standard deviation allowed for y_true.
    ) -> np.ndarray: # R-squared (coefficient of determination) for each channel.
    """
    Computes the R-squared (coefficient of determination) score.

    Calculates R^2 = 1 - NRMSE^2 separately for each channel.
    Returns NaN for channels where NRMSE could not be calculated (e.g., std(y_true) near zero).
    A constant model that always predicts the mean of y_true would get R^2=0.
    """
    nrmse_val = nrmse(y_true, y_pred, time_axis=time_axis, std_tolerance=std_tolerance)

    # R^2 calculation, handles potential NaNs from nrmse
    # Ensure we don't take the square of NaN resulting in NaN where 0 might be expected if rmse is 0
    r2 = 1.0 - nrmse_val**2

    return r2



"""Fill in a module description here"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/datasets/industrial_robot.ipynb.

# %% auto 0
__all__ = ['u_forward', 'y_forward', 'BenchmarkRobotForward_Simulation', 'BenchmarkRobotForward_Prediction', 'u_inverse',
           'y_inverse', 'BenchmarkRobotInverse_Simulation', 'BenchmarkRobotInverse_Prediction', 'dl_robot_forward',
           'dl_robot_inverse']

# %% ../../nbs/datasets/industrial_robot.ipynb 2
from ..utils import write_dataset,write_array
import identibench.benchmark as idb
import identibench.metrics
from nonlinear_benchmarks.utilities import cashed_download
from pathlib import Path
import os
import h5py
import numpy as np
import scipy.io as sio

# %% ../../nbs/datasets/industrial_robot.ipynb 3
def robot_mat2hdf(
        save_path: Path, #directory the files are written to, created if it does not exist
        mat_path: Path, #path of mat file to extract
) -> None:         
    'converts .mat file of industrial robot to hdf5 file, used for forward and inverse'


    fs = 10  # Hz
    train_valid_split = 0.8
    os.makedirs(save_path / 'test', exist_ok=True)
    os.makedirs(save_path / 'train', exist_ok=True)
    os.makedirs(save_path / 'valid', exist_ok=True)

    mf = sio.loadmat(mat_path)
    for mode in ['train', 'test']:
        if mode == 'test':
            with h5py.File(save_path / 'test' / 'test.hdf5', 'w') as f:
                write_dataset(f, 'dt', np.ones_like(mf[f'time_{mode}'][0]) / fs)
                write_array(f, 'u', mf[f'u_{mode}'].T)
                write_array(f, 'y', mf[f'y_{mode}'].T)
                f.attrs['fs'] = fs

        else:
            with h5py.File(save_path / 'train' / 'train.hdf5', 'w') as train_f, \
                h5py.File(save_path / 'valid' / 'valid.hdf5', 'w') as valid_f:
                    dt = np.ones_like(mf[f'time_{mode}'][0]) / fs
                    total_entries = len(dt)
                    split_index = int(total_entries * train_valid_split)

                    write_dataset(train_f, 'dt', dt[:split_index])
                    write_array(train_f, 'u', mf[f'u_{mode}'][:,:split_index].T)
                    write_array(train_f, 'y', mf[f'y_{mode}'][:,:split_index].T)
                    train_f.attrs['fs'] = fs  
                    
                    write_dataset(valid_f, 'dt', dt[split_index:])
                    write_array(valid_f, 'u', mf[f'u_{mode}'][:,split_index:].T)
                    write_array(valid_f, 'y', mf[f'y_{mode}'][:,split_index:].T)
                    valid_f.attrs['fs'] = fs  
    

# %% ../../nbs/datasets/industrial_robot.ipynb 4
def dl_robot_forward(
        save_path: Path, #directory the files are written to, created if it does not exist
        force_download: bool = False # force download the dataset
) -> None:
    save_path = Path(save_path)
    url_robot = "https://fdm-fallback.uni-kl.de/TUK/FB/MV/WSKL/0001/Robot_Identification_Benchmark_Without_Raw_Data.rar"
    # unrar_download(url_robot,tmp_dir)

    tmp_dir = cashed_download(url_robot,'Industrial_robot',force_download=force_download)
    tmp_dir = Path(tmp_dir)

    path_forward = tmp_dir / "forward_identification_without_raw_data.mat"

    robot_mat2hdf(save_path,path_forward)
    

# %% ../../nbs/datasets/industrial_robot.ipynb 6
u_forward = [ f'u{i}' for i in range(0,6) ]
y_forward = [ f'y{i}' for i in range(0,6) ]

BenchmarkRobotForward_Simulation = idb.BenchmarkSpecSimulation(
    name='BenchmarkRobotForward_Simulation', dataset_id='robot_forward',
    u_cols=u_forward, y_cols=y_forward, metric_func=identibench.metrics.rmse, 
    download_func=dl_robot_forward,
    init_window=100
)
BenchmarkRobotForward_Prediction = idb.BenchmarkSpecPrediction(
    name='BenchmarkRobotForward_Prediction', dataset_id='robot_forward',
    u_cols=u_forward, y_cols=y_forward, metric_func=identibench.metrics.rmse, 
    download_func=dl_robot_forward,
    init_window=100, pred_horizon=100,pred_step=100
)

# %% ../../nbs/datasets/industrial_robot.ipynb 9
def dl_robot_inverse(
        save_path: Path, #directory the files are written to, created if it does not exist
        force_download: bool = False # force download the dataset
) -> None:
    save_path = Path(save_path)
    url_robot = "https://fdm-fallback.uni-kl.de/TUK/FB/MV/WSKL/0001/Robot_Identification_Benchmark_Without_Raw_Data.rar"
    # unrar_download(url_robot,tmp_dir)

    tmp_dir = cashed_download(url_robot,'Industrial_robot',force_download=force_download)
    tmp_dir = Path(tmp_dir)

    path_inverse = tmp_dir / "inverse_identification_without_raw_data.mat"

    robot_mat2hdf(save_path,path_inverse)
    

# %% ../../nbs/datasets/industrial_robot.ipynb 11
u_inverse = [ f'u{i}' for i in range(0,12) ]
y_inverse = [ f'y{i}' for i in range(0,6) ]

BenchmarkRobotInverse_Simulation = idb.BenchmarkSpecSimulation(
    name='BenchmarkRobotInverse_Simulation', dataset_id='robot_inverse',
    u_cols=u_inverse, y_cols=y_inverse, metric_func=identibench.metrics.rmse, 
    download_func=dl_robot_inverse,
    init_window=100
)
BenchmarkRobotInverse_Prediction = idb.BenchmarkSpecPrediction(
    name='BenchmarkRobotInverse_Prediction', dataset_id='robot_inverse',
    u_cols=u_inverse, y_cols=y_inverse, metric_func=identibench.metrics.rmse, 
    download_func=dl_robot_inverse,
    init_window=100, pred_horizon=100,pred_step=100
)
